@@ -1,7 +1,9 @@
+import codecs
 import collections
 import contextlib
-import re
+import csv
 from datetime import datetime
+from grpc._channel import _InactiveRpcError
 import logging
 import numpy
 import os
@@ -10,12 +12,13 @@ from pandas.errors import EmptyDataError
 import prometheus_client
 import string
 import random
+import re
 import warnings
 
 from .numerical import mean_stddev, get_numerical_ranges
 from .profile_types import identify_types, determine_dataset_type
-from .spatial import nominatim_resolve_all, pair_latlong_columns, \
-    get_spatial_ranges, parse_wkt_column
+from .spatial import LatLongColumn, nominatim_resolve_all, \
+    pair_latlong_columns, get_spatial_ranges, parse_wkt_column
 from .temporal import get_temporal_resolution
 from . import types
 
@@ -30,8 +33,6 @@ SAMPLE_ROWS = 20
 
 MAX_UNCLEAN_ADDRESSES = 0.20  # 20%
 
-MAX_WRONG_LEVEL_ADMIN = 0.10  # 10%
-
 
 BUCKETS = [0.5, 1.0, 5.0, 10.0, 20.0, 30.0, 60.0, 120.0, 300.0, 600.0]
 
@@ -95,6 +96,14 @@ def expand_attribute_name(name):
     yield ''.join(word)
 
 
+def _lazo_retry(func):
+    try:
+        return func()
+    except _InactiveRpcError:
+        pass
+    return func()
+
+
 @PROM_PROFILE.time()
 def process_dataset(data, dataset_id=None, metadata=None,
                     lazo_client=None, nominatim=None, geo_data=None,
@@ -146,8 +155,11 @@ def process_dataset(data, dataset_id=None, metadata=None,
         # https://github.com/pandas-dev/pandas/issues/25353 (nan as str 'nan')
         data = data.astype(object).fillna('').astype(str)
 
+        column_names = data.columns
+
         # FIXME: no sampling here!
     else:
+        column_names = None  # Avoids a warning
         with contextlib.ExitStack() as stack:
             if isinstance(data, (str, bytes)):
                 if not os.path.exists(data):
@@ -170,6 +182,27 @@ def process_dataset(data, dataset_id=None, metadata=None,
                 raise TypeError("data should be a filename, a file object, or "
                                 "a pandas.DataFrame")
 
+            # Read column names
+            read_sample = data.read(4)
+            data.seek(0, 0)
+            if isinstance(read_sample, str):
+                reader = csv.reader(data)
+                try:
+                    column_names = next(reader)
+                except StopIteration:
+                    column_names = None
+                del reader
+            else:
+                codec_reader = codecs.getreader('utf-8')(data)
+                reader = csv.reader(codec_reader)
+                try:
+                    column_names = next(reader)
+                except StopIteration:
+                    column_names = None
+                del reader
+                del codec_reader
+            data.seek(0, 0)
+
             # Load the data
             try:
                 if metadata['size'] > load_max_size:
@@ -204,20 +237,19 @@ def process_dataset(data, dataset_id=None, metadata=None,
 
     metadata['nb_profiled_rows'] = data.shape[0]
 
-    # Get column dictionary
-    columns = metadata.setdefault('columns', [])
-    # Fix size if wrong
-    if len(columns) != len(data.columns):
-        logger.info("Setting column names from header")
-        columns[:] = [{} for _ in data.columns]
+    if 'columns' in metadata:
+        columns = metadata['columns']
+        logger.info("Using provided columns info")
+        if len(columns) != len(data.columns):
+            raise ValueError("Column metadata doesn't match number of columns")
+        for column_meta, name in zip(columns, column_names):
+            if 'name' in column_meta and column_meta['name'] != name:
+                raise ValueError("Column names don't match")
+            column_meta['name'] = name
     else:
-        logger.info("Keeping columns from discoverer")
-
-    # Set column names
-    for column_meta, name in zip(columns, data.columns):
-        if 'name' in column_meta and column_meta['name'] != name:
-            raise ValueError("Column names don't match")
-        column_meta['name'] = name
+        logger.info("Setting column names from header")
+        columns = [{'name': name} for name in column_names]
+        metadata['columns'] = columns
 
     if data.shape[0] == 0:
         logger.info("0 rows, returning early")
@@ -324,17 +356,23 @@ def process_dataset(data, dataset_id=None, metadata=None,
 
                 # Get lat/long columns
                 if types.LATITUDE in semantic_types_dict:
-                    columns_lat.append((
-                        column_meta['name'],
-                        numerical_values,
-                        manual_latlong_pairs.get(column_meta['name']),
-                    ))
+                    columns_lat.append(
+                        LatLongColumn(
+                            index=column_idx,
+                            name=column_meta['name'],
+                            values=numerical_values,
+                            annot_pair=manual_latlong_pairs.get(column_meta['name']),
+                        )
+                    )
                 elif types.LONGITUDE in semantic_types_dict:
-                    columns_long.append((
-                        column_meta['name'],
-                        numerical_values,
-                        manual_latlong_pairs.get(column_meta['name']),
-                    ))
+                    columns_long.append(
+                        LatLongColumn(
+                            index=column_idx,
+                            name=column_meta['name'],
+                            values=numerical_values,
+                            annot_pair=manual_latlong_pairs.get(column_meta['name']),
+                        )
+                    )
                 elif coverage:
                     ranges = get_numerical_ranges(
                         [x for x in numerical_values if x is not None]
@@ -448,24 +486,16 @@ def process_dataset(data, dataset_id=None, metadata=None,
                 if non_empty > 0:
                     unclean_ratio = 1.0 - len(locations) / non_empty
                     if unclean_ratio <= MAX_UNCLEAN_ADDRESSES:
-                        resolved_addresses[column_meta['name']] = locations
+                        resolved_addresses[column_idx] = locations
                         if types.ADDRESS not in column_meta['semantic_types']:
                             column_meta['semantic_types'].append(types.ADDRESS)
 
             # Guess level of administrative areas
             if types.ADMIN in semantic_types_dict:
-                areas = semantic_types_dict[types.ADMIN]
-                level_counter = collections.Counter()
-                for area in areas:
-                    if area is not None:
-                        level_counter[area.level] += 1
-                threshold = (1.0 - MAX_WRONG_LEVEL_ADMIN) * len(areas)
-                threshold = max(3, threshold)
-                for level, count in level_counter.items():
-                    if count >= threshold:
-                        column_meta['admin_area_level'] = level
-                        break
-                resolved_admin_areas[column_meta['name']] = areas
+                level, areas = semantic_types_dict[types.ADMIN]
+                if level is not None:
+                    column_meta['admin_area_level'] = level
+                resolved_admin_areas[column_idx] = areas
 
     # Textual columns
     if lazo_client and columns_textual:
@@ -475,21 +505,25 @@ def process_dataset(data, dataset_id=None, metadata=None,
             try:
                 if data_path:
                     # if we have the path, send the path
-                    lazo_client.index_data_path(
-                        data_path,
-                        dataset_id,
-                        [columns[idx]['name'] for idx in columns_textual],
-                    )
+                    def call_lazo():
+                        lazo_client.index_data_path(
+                            data_path,
+                            dataset_id,
+                            [columns[idx]['name'] for idx in columns_textual],
+                        )
+                    _lazo_retry(call_lazo)
                 else:
                     # if path is not available, send the data instead
                     for idx in columns_textual:
-                        lazo_client.index_data(
-                            data.iloc[:, idx].values.tolist(),
-                            dataset_id,
-                            columns[idx]['name'],
-                        )
+                        def call_lazo():
+                            lazo_client.index_data(
+                                data.iloc[:, idx].values.tolist(),
+                                dataset_id,
+                                columns[idx]['name'],
+                            )
+                        _lazo_retry(call_lazo)
             except Exception:
-                logger.error("Error indexing textual attributes from %s", dataset_id)
+                logger.warning("Error indexing textual attributes from %s", dataset_id)
                 raise
         # Generating Lazo sketches for the search
         else:
@@ -497,22 +531,24 @@ def process_dataset(data, dataset_id=None, metadata=None,
             try:
                 if data_path:
                     # if we have the path, send the path
-                    lazo_sketches = lazo_client.get_lazo_sketch_from_data_path(
-                        data_path,
-                        "",
-                        [columns[idx]['name'] for idx in columns_textual],
-                    )
+                    def call_lazo():
+                        return lazo_client.get_lazo_sketch_from_data_path(
+                            data_path,
+                            "",
+                            [columns[idx]['name'] for idx in columns_textual],
+                        )
+                    lazo_sketches = _lazo_retry(call_lazo)
                 else:
                     # if path is not available, send the data instead
                     lazo_sketches = []
                     for idx in columns_textual:
-                        lazo_sketches.append(
-                            lazo_client.get_lazo_sketch_from_data(
+                        def call_lazo():
+                            return lazo_client.get_lazo_sketch_from_data(
                                 data.iloc[:, idx].values.tolist(),
                                 "",
                                 columns[idx]['name'],
                             )
-                        )
+                        lazo_sketches.append(_lazo_retry(call_lazo))
                 # saving sketches into metadata
                 for sketch, idx in zip(lazo_sketches, columns_textual):
                     n_permutations, hash_values, cardinality = sketch
@@ -525,7 +561,6 @@ def process_dataset(data, dataset_id=None, metadata=None,
                 logger.warning("Error getting Lazo sketches")
                 raise
 
-    # Lat / Long
     if coverage:
         logger.info("Computing spatial coverage...")
         spatial_coverage = []
@@ -548,9 +583,9 @@ def process_dataset(data, dataset_id=None, metadata=None,
                     col['semantic_types'].remove(types.LONGITUDE)
 
             # Compute ranges from lat/long pairs
-            for (name_lat, values_lat, annot_pair), (name_long, values_long, annot_pair) in pairs:
+            for col_lat, col_long in pairs:
                 values = []
-                for lat, long in zip(values_lat, values_long):
+                for lat, long in zip(col_lat.values, col_long.values):
                     if (lat and long and  # Ignore None and 0
                             -90 < lat < 90 and -180 < long < 180):
                         values.append((lat, long))
@@ -558,13 +593,19 @@ def process_dataset(data, dataset_id=None, metadata=None,
                 if len(values) > 1:
                     logger.info(
                         "Computing spatial ranges lat=%r long=%r (%d rows)",
-                        name_lat, name_long, len(values),
+                        col_lat.name, col_long.name, len(values),
                     )
                     spatial_ranges = get_spatial_ranges(values)
                     if spatial_ranges:
-                        spatial_coverage.append({"lat": name_lat,
-                                                 "lon": name_long,
-                                                 "ranges": spatial_ranges})
+                        spatial_coverage.append({
+                            'type': 'latlong',
+                            'column_names': [col_lat.name, col_long.name],
+                            'column_indexes': [
+                                col_lat.index,
+                                col_long.index,
+                            ],
+                            'ranges': spatial_ranges,
+                        })
 
             # Compute ranges from WKT points
             for i, col in enumerate(columns):
@@ -578,27 +619,36 @@ def process_dataset(data, dataset_id=None, metadata=None,
                 )
                 spatial_ranges = get_spatial_ranges(values)
                 if spatial_ranges:
-                    spatial_coverage.append({"point": name,
-                                             "ranges": spatial_ranges})
+                    spatial_coverage.append({
+                        'type': 'point',
+                        'column_names': [name],
+                        'column_indexes': [i],
+                        'ranges': spatial_ranges,
+                    })
 
             # Compute ranges from addresses
-            for name, values in resolved_addresses.items():
+            for idx, values in resolved_addresses.items():
+                name = columns[idx]['name']
                 logger.info(
                     "Computing spatial ranges address=%r (%d rows)",
                     name, len(values),
                 )
                 spatial_ranges = get_spatial_ranges(values)
                 if spatial_ranges:
-                    spatial_coverage.append({"address": name,
-                                             "ranges": spatial_ranges})
+                    spatial_coverage.append({
+                        'type': 'address',
+                        'column_names': [name],
+                        'column_indexes': [idx],
+                        'ranges': spatial_ranges,
+                    })
 
             # Compute ranges from administrative areas
-            for name, areas in resolved_admin_areas.items():
+            for idx, areas in resolved_admin_areas.items():
                 merged = None
                 for area in areas:
                     if area is None:
                         continue
-                    new = geo_data.get_bounds(area.area)
+                    new = area.bounds
                     if new:
                         if merged is None:
                             merged = new
@@ -615,8 +665,10 @@ def process_dataset(data, dataset_id=None, metadata=None,
                     and merged[3] - merged[2] > 0.01
                 ):
                     spatial_coverage.append({
-                        "admin": name,
-                        "ranges": [
+                        'type': 'admin',
+                        'column_names': [columns[idx]['name']],
+                        'column_indexes': [idx],
+                        'ranges': [
                             {
                                 'range': {
                                     'type': 'envelope',
