@@ -51,7 +51,7 @@ _re_other_point = re.compile(
     r'-?[0-9]{1,3}\.[0-9]{1,15}'
     r'\)$'
 )
-_re_whitespace = re.compile(r'\s')
+_re_whitespace = re.compile(r'\s+')
 
 
 # Tolerable ratio of unclean data
@@ -62,6 +62,9 @@ MAX_UNCLEAN = 0.02  # 2%
 MAX_CATEGORICAL_RATIO = 0.10  # 10%
 
 
+MAX_WRONG_LEVEL_ADMIN = 0.10  # 10%
+
+
 def regular_exp_count(array):
     # Let you check/count how many instances match a structure of a data type
     re_count = collections.Counter()
@@ -81,7 +84,7 @@ def regular_exp_count(array):
             re_count['other_point'] += 1
         elif _re_wkt_polygon.match(elem):
             re_count['polygon'] += 1
-        elif len(_re_whitespace.findall(elem)) >= 4:
+        elif len(_re_whitespace.findall(elem)) >= 3:
             re_count['text'] += 1
         if elem.lower() in ('0', '1', 'true', 'false', 'y', 'n', 'yes', 'no'):
             re_count['bool'] += 1
@@ -129,14 +132,34 @@ def identify_structural_type(re_count, num_total, threshold):
         structural_type = types.GEO_POINT
     elif re_count['polygon'] >= threshold:
         structural_type = types.GEO_POLYGON
-    elif re_count['polygon'] >= threshold:
-        structural_type = types.GEO_POLYGON
     else:
         structural_type = types.TEXT
 
     return structural_type
 
 
+def guess_admin_level(admin_areas):
+    level_counter = collections.Counter()
+    # `area` is a list of lists of areas
+    # For each name in the original data, it contains a list of the
+    # areas that were found with that name
+    for areas_resolved in admin_areas:
+        # Count each possible admin level only once
+        levels = set(
+            area.type.value
+            for area in areas_resolved
+            if 0 <= area.type.value <= 5
+        )
+        level_counter.update(levels)
+    threshold = (1.0 - MAX_WRONG_LEVEL_ADMIN) * len(admin_areas)
+    threshold = max(3, threshold)
+    for level, count in sorted(level_counter.items()):
+        if count >= threshold:
+            return level
+    else:
+        return None
+
+
 def identify_types(array, name, geo_data, manual=None):
     num_total = len(array)
     column_meta = {}
@@ -173,9 +196,16 @@ def identify_types(array, name, geo_data, manual=None):
                 semantic_types_dict[types.DATE_TIME] = dates
             if el == types.ADMIN:
                 if geo_data is not None:
-                    resolved = geo_data.resolve_names(array)
-                    if sum(1 for r in resolved if r is not None) > 0.7 * len(array):
-                        semantic_types_dict[types.ADMIN] = resolved
+                    resolved = geo_data.resolve_names_all(array)
+                    resolved = [r for r in resolved if r]
+                    if resolved:
+                        level = guess_admin_level(resolved)
+                        if level is not None:
+                            resolved = [
+                                area for areas_list in resolved for area in areas_list
+                                if area.type.value == level
+                            ]
+                            semantic_types_dict[types.ADMIN] = level, resolved
             if el == types.CATEGORICAL or el == types.INTEGER:
                 # Count distinct values
                 values = set(e for e in array if e)
@@ -196,11 +226,19 @@ def identify_types(array, name, geo_data, manual=None):
         if structural_type == types.TEXT:
             categorical = False
 
+            # Administrative areas
             if geo_data is not None:
-                resolved = geo_data.resolve_names(array)
-                if sum(1 for r in resolved if r is not None) > 0.7 * len(array):
-                    semantic_types_dict[types.ADMIN] = resolved
-                    categorical = True
+                resolved = geo_data.resolve_names_all(array)
+                resolved = [r for r in resolved if r]
+                if len(resolved) > 0.7 * len(array):
+                    level = guess_admin_level(resolved)
+                    if level is not None:
+                        resolved = [
+                            area for areas_list in resolved for area in areas_list
+                            if area.type.value == level
+                        ]
+                        semantic_types_dict[types.ADMIN] = level, resolved
+                        categorical = True
 
             if not categorical and num_text >= threshold:
                 # Free text
